@@ -31,34 +31,39 @@
 #include "main.h"
 #include "error.h"
 #include <ft2build.h>
-#include FT_FREETYPE_H
-#include "hint.h"
-#include "hfonts.h"
-#include "hrender.h"
+#include <freetype/freetype.h>
 #include "rendernative.h"
+#define STBI_ONLY_JPEG
+#define STBI_ONLY_PNG
+#define STBI_ONLY_BMP
 #define STB_IMAGE_IMPLEMENTATION
 #include "stb_image.h"
 
 
 #ifdef DEBUG
-static void printGLString(const char *name, GLenum s) {
-    const char *v = (const char *) glGetString(s);
-    LOG("GL %s = %s\n", name, v);
-}
 
 static void checkGlError(const char *op)
 { GLint error;
   while( (error= glGetError())!= GL_NO_ERROR)
 	  MESSAGE("OGL Error after %s: 0x%x\n", op, error);
 }
+
+static void printGLString(const char *name, GLenum s) {
+  const GLubyte *v;
+  v = glGetString(s);
+  checkGlError(name);
+  LOG("GL %s = %s\n", name, v);
+}
+
 #else
 #define printGLString(N,S)	(void)0
 #define checkGlError(OP)	(void)0
 #endif
 
 #define MAX_INFOLOG 512
-static 	GLuint ProgramID, MatrixID, RuleID, GammaID, FGcolorID, IsImageID, ImageID=0;
+static GLuint ProgramID, MatrixID, RuleID, GammaID, FGcolorID, IsImageID, ImageID=0;
 static unsigned char *last_b=NULL;
+static uint32_t cur_fg=0; /*the current foreground color*/
 
 #define xyID 0
 #define uvID 1
@@ -90,17 +95,19 @@ static const char FragmentShader[]=
   "out vec4 color;\n"
   /* Constants for the current triangles */
   "uniform sampler2D theTexture;\n"
-  "uniform vec3 FGcolor;\n"
+  "uniform vec4 FGcolor;\n"
   "uniform float Gamma;\n"
   "uniform int IsImage;\n"
 
   "void main()\n"
   "{ vec4 texColor = texture2D( theTexture, UV );\n"
-    "if (IsImage==1) color = texColor;\n"
-    "else\n"
-    "{ color.a = pow(texColor.r,Gamma);\n"
-      "color.rgb = FGcolor;\n"
+    "if (IsImage==0) {\n"
+      "color.a = pow(texColor.r*FGcolor.a,Gamma);\n"
+      "color.r = FGcolor.r;\n"
+      "color.g = FGcolor.g;\n"
+      "color.b = FGcolor.b;\n"
     "}\n"
+    "else color = texColor;\n"
   "}\n"
 ;
 
@@ -132,13 +139,17 @@ static GLuint loadShader(GLenum type, char const *source)
 
   /* Compile and check vertex shader */
   shaderID = glCreateShader(type);
-  glShaderSource(shaderID, 1, &source , NULL);
-  glCompileShader(shaderID);
-  glGetShaderiv(shaderID, GL_COMPILE_STATUS, &result);
-  if (!result)
-  { char InfoLog[MAX_INFOLOG];
-    glGetShaderInfoLog(shaderID, MAX_INFOLOG, NULL, InfoLog);
-    QUIT("Error compiling shader (%d): %s\n", type, InfoLog);
+  if (shaderID) {
+    glShaderSource(shaderID, 1, &source, NULL);
+    glCompileShader(shaderID);
+    glGetShaderiv(shaderID, GL_COMPILE_STATUS, &result);
+    if (!result)
+    { char InfoLog[MAX_INFOLOG];
+      glGetShaderInfoLog(shaderID, MAX_INFOLOG, NULL, InfoLog);
+      glDeleteShader(shaderID);
+      shaderID = 0;
+      QUIT("Error compiling shader (%d): %s\n", type, InfoLog);
+    }
   }
   return shaderID;
 }
@@ -220,7 +231,7 @@ static void mkRuleTexture() /* the texture used for rules */
   rule[1][1][1]=0xFF; // green bottom right
 #endif
 	// Create one OpenGL texture
-  glPixelStorei(GL_UNPACK_ALIGNMENT, 1); // select byte alignment
+  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
   checkGlError("glPixelStorei");
 
   glGenTextures(1, &RuleID);
@@ -238,11 +249,8 @@ static void mkRuleTexture() /* the texture used for rules */
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
-
-  //LOG("mkRuleTexture Done\n");
 }
 
-
 void nativeInit(void)
 { GLuint VertexArrayID;
   static int glewInitialized=0;
@@ -253,22 +261,43 @@ void nativeInit(void)
   glewInitialized=1;
   while(glGetError()!= GL_NO_ERROR)
     continue; /* ignore errors in glewInit */
+#ifdef DEBUG
+  { int a=0;
+    glGetFramebufferAttachmentParameteriv(GL_FRAMEBUFFER,GL_FRONT_LEFT,
+					  GL_FRAMEBUFFER_ATTACHMENT_OBJECT_TYPE,&a);
+     fprintf(stderr,"Type = %d\n",a);
+
+
+    
+    glGetFramebufferAttachmentParameteriv(GL_FRAMEBUFFER,GL_FRONT_LEFT,
+					  GL_FRAMEBUFFER_ATTACHMENT_RED_SIZE,&a);
+     fprintf(stderr,"Red Size = %d\n",a);
+    glGetFramebufferAttachmentParameteriv(GL_FRAMEBUFFER,GL_FRONT_LEFT,
+					  GL_FRAMEBUFFER_ATTACHMENT_GREEN_SIZE,&a);
+     fprintf(stderr,"Green Size = %d\n",a);
+    glGetFramebufferAttachmentParameteriv(GL_FRAMEBUFFER,GL_FRONT_LEFT,
+					  GL_FRAMEBUFFER_ATTACHMENT_BLUE_SIZE,&a);
+     fprintf(stderr,"Blue Size = %d\n",a);
+    glGetFramebufferAttachmentParameteriv(GL_FRAMEBUFFER,GL_FRONT_LEFT,
+					  GL_FRAMEBUFFER_ATTACHMENT_ALPHA_SIZE,&a);
+     fprintf(stderr,"Alpha Size = %d\n",a);
+  }
 
   printGLString("Version", GL_VERSION);
   printGLString("Vendor", GL_VENDOR);
   printGLString("Renderer", GL_RENDERER);
-  printGLString("Extensions", GL_EXTENSIONS);
-
+#endif
     /* Create the vertex array object (VAO) */
   glGenVertexArrays(1, &VertexArrayID);
+  checkGlError("glGenVertexArrays");
   glBindVertexArray(VertexArrayID);
-
+  checkGlError("glBindVertexArray");
   createProgram();
 
   MatrixID = glGetUniformLocation(ProgramID, "MVP");
-  FGcolorID  = glGetUniformLocation(ProgramID, "FGcolor");
+  FGcolorID = glGetUniformLocation(ProgramID, "FGcolor");
   GammaID  = glGetUniformLocation(ProgramID, "Gamma");
-  IsImageID  = glGetUniformLocation(ProgramID, "IsImage");
+  IsImageID = glGetUniformLocation(ProgramID, "IsImage");
   checkGlError("get IDs"); 
 
   glGenBuffers(1, &xybuffer);
@@ -285,9 +314,9 @@ void nativeInit(void)
   glUniformMatrix4fv(MatrixID, 1, GL_FALSE, &MVP[0][0]);
   glUniform1f(GammaID, 1.0f/2.2f);
   glUniform1i(IsImageID, 0);
-  glUniform3f(FGcolorID, 0.0, 0.0, 0.0); // black as default foreground
-  glClearColor(1.0f, 1.0f, 1.0f, 0.0f); // white as default background
-
+  glUniform4f(FGcolorID, 0.0, 0.0, 0.0, 1.0); // black as default foreground
+  cur_fg=0x000000FF; /* possibly better: use nativeSetForeground */
+  glClearColor(1.0f, 1.0f, 1.0f, 1.0f); // white as default background
    /* 1rst attribute buffer : XYs */
   glEnableVertexAttribArray(xyID);
   glBindBuffer(GL_ARRAY_BUFFER, xybuffer);
@@ -312,11 +341,10 @@ void nativeInit(void)
   glEnable(GL_FRAMEBUFFER_SRGB);
   checkGlError("GL_FRAMEBUFFER_SRGB");
 
-  hint_clear_fonts(true);
   mkRuleTexture();
   ImageID=0;
   last_b=NULL;
-  //LOG("nativeInit Done\n");
+  //LOG("nativeInit done\n");
 }
 
 
@@ -331,55 +359,46 @@ void nativeClear(void)
   }
 }
 
-void nativeBlank(void)
-{ glClear(GL_COLOR_BUFFER_BIT);
-}
-
 void nativeSetGamma(double gamma)
 { glUniform1f(GammaID, 1.0/gamma);
   checkGlError("glsetgamma");
 }
 
-static GLfloat curfr=0.0f, curfg=0.0f, curfb=0.0f;
-static uint8_t last_style=0;
-static void nativeSetColors(GLfloat fr, GLfloat fg, GLfloat fb, GLfloat br, GLfloat bg, GLfloat bb)
-/* set foreground and background rgb colors */
+
+
+void nativeSetForeground(uint32_t fg)
+/* set foreground rgba colors */
+{ if (fg!=cur_fg)
+  { uint8_t r,g,b,a;
+    cur_fg=fg;
+    a=fg&0xFF;fg=fg>>8;
+    b=fg&0xFF;fg=fg>>8;
+    g=fg&0xFF;fg=fg>>8;
+    r=fg&0xFF;
+    glUniform4f(FGcolorID, r/255.0f,g/255.0f,b/255.0f,a/255.0f);
+  }
+}
+
+
+void nativeBackground(double x, double y, double w, double h, uint32_t bg)
+{ uint32_t fg=cur_fg;
+  nativeSetForeground(bg);
+  nativeRule(x,y,w,h);
+  nativeSetForeground(fg);
+}
+
+void nativeBlank(uint32_t bg)
+{ uint8_t r,g,b,a;
+  a=bg&0xFF;bg=bg>>8;
+  b=bg&0xFF;bg=bg>>8;
+  g=bg&0xFF;bg=bg>>8;
+  r=bg&0xFF;
+  glClearColor(	r/255.0f,g/255.0f,b/255.0f,a/255.0f);
+  glClear(GL_COLOR_BUFFER_BIT);
+}
+
+void nativeSetSize(int px_h, int px_v, double pt_h, double pt_v)
 {
-  glClearColor(br, bg, bb, 1.0f);
-  curfr=fr; curfg=fg; curfb=fb;
-  glUniform3f(FGcolorID, fr, fg, fb);
-  last_style=0;
-}
-
-
-void nativeSetDark(int on)
-{   if (on) {
-        nativeSetColors(
-			GET_R(FG_NIGHT)/255.0f, 
-			GET_G(FG_NIGHT)/255.0f, 
-			GET_B(FG_NIGHT)/255.0f, 
-			GET_R(BG_NIGHT)/255.0f, 
-			GET_G(BG_NIGHT)/255.0f, 
-			GET_B(BG_NIGHT)/255.0f);
-    } else {
-        nativeSetColors(
-			GET_R(FG_DAY)/255.0f, 
-			GET_G(FG_DAY)/255.0f, 
-			GET_B(FG_DAY)/255.0f, 
-			GET_R(BG_DAY)/255.0f, 
-			GET_G(BG_DAY)/255.0f, 
-			GET_B(BG_DAY)/255.0f);
-     }
-}
-
-static float pt_h=600.0, pt_v=800.0;
-
-void nativeSetSize(int px_h, int px_v, double x_dpi, double y_dpi)
-{  //LOG("xdpi=%f, ydpi=%f\n",x_dpi,y_dpi);
-   pt_h = px_h * 72.27 / x_dpi;
-   pt_v = px_v * 72.27 / y_dpi; 
-   page_h = round(pt_h * (1 << 16));
-   page_v = round(pt_v * (1 << 16));
    MVP[0][0]=2.0/pt_h; // x: scale to -1 to +1
    MVP[1][1]=-2.0/pt_v; // y: scale to 1 to -1
    //MVP[2][2]=0.0f; // z: don't care
@@ -418,9 +437,9 @@ void nativeImage(double x, double y, double w, double h, unsigned char *b, unsig
 /* render the image found between *b and *e at x,y with size w,h.
    x, y, w, h are given in point
 */
-{
-  GLenum format, internal_format;
+{ GLenum format;
   int width, height, nrChannels;
+
   if (b!=last_b||ImageID==0)
   { unsigned char *data;
     static unsigned char grey[4]={0,0x80,0x80,0x80};
@@ -434,27 +453,30 @@ void nativeImage(double x, double y, double w, double h, unsigned char *b, unsig
     { LOG("Unable to display image\n");
       data=grey; width=height=1; nrChannels=4;
     }
-    //LOG("nativeImage %d chanels\n",nrChannels);	
-    internal_format=GL_SRGB;
+    //LOG("nativeImage %d chanels\n",nrChannels);
+    format = GL_RGBA;
     if (nrChannels == 4)
-    { format = GL_RGBA; internal_format=GL_SRGB_ALPHA;}
-    else if (nrChannels == 3) 
-      format = GL_RGB;
-    else if (nrChannels == 2) 
-      format = GL_RG;
-    else
-      format = GL_RED;
+        format = GL_RGBA;
+    else if (nrChannels == 3)
+        format = GL_RGB;
+    else if (nrChannels == 2)
+        format = GL_LUMINANCE_ALPHA;
+    else if (nrChannels == 1)
+        format = GL_LUMINANCE;
     glGenTextures(1, &ImageID);
     glBindTexture(GL_TEXTURE_2D, ImageID);
     checkGlError("glBindTexture ImageID");
-
-    glTexImage2D(GL_TEXTURE_2D, 0, internal_format, width, height, 0, format, GL_UNSIGNED_BYTE, data);
+    glTexImage2D(GL_TEXTURE_2D, 0, format, width, height, 0,
+                 format, GL_UNSIGNED_BYTE, data);
+    if (glGetError()!= GL_NO_ERROR)
+      glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0,
+    		  format, GL_UNSIGNED_BYTE, data);
     checkGlError("glTexImage2D(image)");
-    if (data!=grey) stbi_image_free(data);
+    if (data!=grey) { stbi_image_free(data); data=NULL; }
     glGenerateMipmap(GL_TEXTURE_2D);
     checkGlError("glGenerateMipmap(image)");
-    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
-    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
+    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
+    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
   }
   else
   {  glBindTexture(GL_TEXTURE_2D, ImageID);
@@ -475,116 +497,81 @@ void nativeImage(double x, double y, double w, double h, unsigned char *b, unsig
   glUniform1i(IsImageID, 1);
   glDrawArrays(GL_TRIANGLES, 0, 2*3);
   glUniform1i(IsImageID, 0);
-
 }
 
+static const int to_nearest=1;
 
-static void GLtexture(Gcache *g) {
-    unsigned texID;
+unsigned int nativeTexture(unsigned char *bits, int w, int h) {
+    unsigned int textureID;
     glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
     checkGlError("glPixelStorei");
-    glGenTextures(1, &texID);
+    glGenTextures(1, &textureID);
     checkGlError("glGenTextures");
-    glBindTexture(GL_TEXTURE_2D, texID);
-    checkGlError("glBindTexture texID");
-	
+    glBindTexture(GL_TEXTURE_2D, textureID);
+    checkGlError("glBindTexture textureID");
+#if 0   /* testing pixel alignment */
+    bits[0]=0xFF;      /* The first element in bits corresponds to the top left pixel.*/
+    bits[2]=0xFF; 
+    bits[4]=0xFF; 
+    bits[6]=0xFF; 
+    bits[8]=0xFF; 
 
-    /* the first element in g->bits corresponds to the lower left pixel,
-     * the last element in g->bits to the upper right pixel. */
-  glTexImage2D(
-	    GL_TEXTURE_2D,
+    bits[w-1]=0xFF;    /* This is the top right pixel. */
+    bits[w*h-1] =0xFF; /* The last element in bits is the bottom right pixel. */
+#endif    
+    glTexImage2D(
+            GL_TEXTURE_2D,
             0,
             GL_R8,
-            g->w,
-            g->h,
+            w,
+            h,
             0,
             GL_RED,
             GL_UNSIGNED_BYTE,
-            g->bits
-  );
-  checkGlError("glTeXImage2D Glyph");
-    
+            bits
+    );
+    checkGlError("glTeXImage2D Glyph");
 
-  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
-  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
-  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
-  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
-
-  g->GLtexture = texID;
-  //MESSAGE("Generated GL texture %d",g->GLtexture);
-}
-
-
-
-void nativeSetPK(struct gcache_s *g)
-/* the bits, w and h fields in g are already set, but the rows needs to be DWORD aligned
-   the bitmap is already bottom up */
-{ GLtexture(g);
-}
-
-void nativeSetFreeType(struct gcache_s *g)
-/* the bits, w and h fields in g are already set, but the rows needs to be DWORD aligned
-   and converted to a bottom up DIB
-*/
-{GLtexture(g);}
-
-int round_to_pixel=1; /* makes sense only if using the native dpi, if using a multiple its of not much use*/
-double pixel_size_threshold= 72.27/200; /*round to pixel only if pixel size in pt is above threshold*/
-void nativeGlyph(double x, double dx, double y, double dy, double w, double h, struct gcache_s *g, uint8_t s)
-/* given glyph g, display g at position x,y in size w,h. x, y, w, h are given in point */
-{  
-	if (g->GLtexture == 0)
-        GLtexture(g);
-	x=x-dx;
-	y=y+h-dy;
-	if (round_to_pixel)
-	{ double pxs;
-	  pxs = 72.27/xdpi; /* pixel size in point */
-	  if (pxs>=pixel_size_threshold)
-	  { x=x/pxs;
-	    x=floor(x+0.5);
-	    x=x*pxs;
-	  }
-	  pxs = 72.27/ydpi; /* pixel size in point */
-	  if (pxs>=pixel_size_threshold)
-	  { y=y/pxs;
-	    y=floor(y+0.5);
-	    y=y*pxs;
-	  }
-	}
-  	xy[0][0]=x;	  xy[0][1]=y;
-	xy[1][0]=x;	  xy[1][1]=y-h;
-	xy[2][0]=x+w;     xy[2][1]=y-h;
-
-	xy[3][0]=x+w;	  xy[3][1]=y;
-	xy[4][0]=x;       xy[4][1]=y;
-	xy[5][0]=x+w;     xy[5][1]=y-h;
-
-    glBindTexture(GL_TEXTURE_2D, g->GLtexture);
-    checkGlError("glBindTexture g->GLtexture");
-    
-	if (s!=last_style)
-	{ if (s&FOCUS_BIT)
-	      glUniform3f(FGcolorID, 0.0, 1.0,0.0); 
-	  else if (s&MARK_BIT)
- 	      glUniform3f(FGcolorID, 1.0, 0.0,0.0); 
-	  else if (s&LINK_BIT)
- 	      glUniform3f(FGcolorID, 0.0, 0.0,1.0); 
-	  else
-  	      glUniform3f(FGcolorID, curfr, curfg,curfb);
-	  last_style=s;
-	}
-    glBindBuffer(GL_ARRAY_BUFFER, xybuffer);
-    glBufferData(GL_ARRAY_BUFFER, sizeof(xy), xy, GL_STREAM_DRAW);
-    glDrawArrays(GL_TRIANGLES, 0, 2*3);
-    //LOG("nativeGlyph %f@%f %fx%f Done\n",x,y,w,h);
-}
-
-void nativeFreeGlyph(struct gcache_s*g)
-{    if (g->GLtexture != 0) {
-        glDeleteTextures(1, &(g->GLtexture)); // probably not needed
-        g->GLtexture = 0;
+    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
+    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
+    if (to_nearest) 
+    { glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
+      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
     }
+    else
+    { glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
+      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
+    }
+    //MESSAGE("Generated GL texture %d",textureID);
+    return textureID;
+}
+
+unsigned int nativeFreeTexture(unsigned int t)
+{ if (t != 0) {
+    glDeleteTextures(1, &t); // probably not needed
+  }
+  return 0;
+}
+
+
+void nativeGlyph(double x, double y, double w, double h, unsigned int t)
+/* display texture t at position x,y in size w, h.
+   x, y, w, h are given in point */
+{
+  xy[0][0]=x;	  xy[0][1]=y;
+  xy[1][0]=x;	  xy[1][1]=y-h;
+  xy[2][0]=x+w;     xy[2][1]=y-h;
+
+  xy[3][0]=x+w;	  xy[3][1]=y;
+  xy[4][0]=x;       xy[4][1]=y;
+  xy[5][0]=x+w;     xy[5][1]=y-h;
+
+  glBindTexture(GL_TEXTURE_2D, t);
+  checkGlError("glBindTexture t");
+  glBindBuffer(GL_ARRAY_BUFFER, xybuffer);
+  glBufferData(GL_ARRAY_BUFFER, sizeof(xy), xy, GL_STREAM_DRAW);
+  glDrawArrays(GL_TRIANGLES, 0, 2*3);
+  //checkGlError("glDrawArrays");
 }
 
 
